"""Facilities for working with colors."""


__version__ = "0.4.0"


__all__ = ["Ansi256", "Color", "ColorPair", "HCL", "Hex", "RGB", "WebColor"]


from .color_pair import ColorPair
from .spaces import HCL, RGB, Ansi256, Color, Hex, WebColor



"""Objects representing colors in different color spaces."""


import math
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Iterable, Iterator, Text, TypeVar, Union

from . import ansi256, colorsys, web

C = TypeVar("C", bound="Color")


class Color(ABC, Iterable[float]):
    """Abstract base class for color spaces."""

    @property
    @abstractmethod
    def rgb(self) -> "RGB":
        """Return the color as an RGB object."""
        raise NotImplementedError()

    @property
    def hex(self) -> "Hex":
        """Return the color as an Hex object."""
        return self.rgb.hex

    @property
    def web_color(self) -> "WebColor":
        """Return the color as a WebColor object."""
        return self.rgb.web_color

    @property
    def ansi256(self) -> "Ansi256":
        """Return the color as an Ansi256 object."""
        return self.rgb.ansi256

    @property
    def hcl(self) -> "HCL":
        """Return the color as an HCL object."""
        return self.rgb.hcl

    def __index__(self) -> int:
        """Return the index of the color as an hexadecimal integer."""
        return colorsys.hex_to_hex(self.hex.hex_code)

    def __eq__(self, other: object) -> bool:
        """Return True if the colors are almost equal in RGB space."""
        if not isinstance(other, Color):
            raise TypeError(f"{other!r} is not a Color")
        return hex(self) == hex(other)

    def __hash__(self) -> int:
        """Return the hash of the color."""
        return hash(hex(self))

    def __iter__(self) -> Iterator[float]:
        """Return an iterator over the color's RGB channels."""
        self_rgb = self.rgb
        yield self_rgb.red
        yield self_rgb.green
        yield self_rgb.blue

    def distance(self, other: "Color") -> float:
        """Return the distance between colors in the HCL color space."""
        self_hcl = self.hcl
        other_hcl = other.hcl
        return math.hypot(
            self_hcl.hue - other_hcl.hue,
            self_hcl.chroma - other_hcl.chroma,
            self_hcl.luminance - other_hcl.luminance,
        )

    def closest(self, colors: Iterable[C]) -> C:
        """Find the color in the given list that is closest to this color."""
        return min(colors, key=self.distance)


@dataclass(frozen=True, eq=False)
class RGB(Color):
    """An RGB color."""

    red: float
    green: float
    blue: float

    N_DIGITS = 2

    def __post_init__(self) -> None:
        """Round the RGB channels."""
        object.__setattr__(self, "red", round(self.red, self.N_DIGITS))
        object.__setattr__(self, "green", round(self.green, self.N_DIGITS))
        object.__setattr__(self, "blue", round(self.blue, self.N_DIGITS))

    @property
    def rgb(self) -> "RGB":
        """Return the color as an RGB object."""
        return self

    @property
    def hex(self) -> "Hex":
        """Return the color as an Hex object."""
        return Hex(colorsys.rgb_to_hex(self.red, self.green, self.blue))

    @property
    def web_color(self) -> "WebColor":
        """Return the color as a WebColor object."""
        return self.closest(map(WebColor, web.colors.keys()))

    @property
    def ansi256(self) -> "Ansi256":
        """Return the color as an Ansi256 object."""
        return self.closest(map(Ansi256, range(len(ansi256.colors))))

    @property
    def hcl(self) -> "HCL":
        """Return the color as an HCL object."""
        return HCL(*colorsys.rgb_to_hcl(self.red, self.green, self.blue))


@dataclass(frozen=True, eq=False)
class Hex(Color):
    """A color represented by a hexadecimal integer."""

    hex_code: Union[int, Text]

    def __repr__(self) -> Text:
        """Return a string representation of the color."""
        if isinstance(self.hex_code, int):
            return f"Hex({self.hex_code:X})"
        return f"Hex({self.hex_code!r})"

    @property
    def rgb(self) -> RGB:
        """Return the color as an RGB object."""
        return RGB(*colorsys.hex_to_rgb(self.hex_code))

    @property
    def hex(self) -> "Hex":
        """Return the color as an Hex object."""
        return self


@dataclass(frozen=True, eq=False)
class WebColor(Color):
    """A color represented by a name."""

    name: Text

    NORM_PATTERN = re.compile(r"[\s\-_]+")

    def __post_init__(self) -> None:
        """Normalize the name of the color."""
        norm_name = self.NORM_PATTERN.sub("", self.name).lower()
        if norm_name not in web.colors:
            raise ValueError(f"{norm_name!r} ({self.name!r}) is not a valid color name")
        object.__setattr__(self, "name", norm_name)

    @property
    def rgb(self) -> RGB:
        """Return the color as an RGB object."""
        return self.hex.rgb

    @property
    def hex(self) -> Hex:
        """Return the color as an Hex object."""
        return Hex(colorsys.web_color_to_hex(self.name))

    @property
    def web_color(self) -> "WebColor":
        """Return the color as a WebColor object."""
        return self


@dataclass(frozen=True, eq=False)
class Ansi256(Color):
    """A color represented by an integer between 0 and 255."""

    code: int

    @property
    def rgb(self) -> RGB:
        """Return the color as an RGB object."""
        return self.hex.rgb

    @property
    def hex(self) -> Hex:
        """Return the color as an Hex object."""
        return Hex(colorsys.ansi256_to_hex(self.code))

    @property
    def ansi256(self) -> "Ansi256":
        """Return the color as an Ansi256 object."""
        return self


@dataclass(frozen=True, eq=False)
class HCL(Color):
    """An HCL color."""

    hue: float
    chroma: float
    luminance: float

    @property
    def rgb(self) -> RGB:
        """Return the color as an RGB object."""
        return RGB(*colorsys.hcl_to_rgb(self.hue, self.chroma, self.luminance))

    @property
    def hcl(self) -> "HCL":
        """Return the color as an HCL object."""
        return self



"""A color pair of foreground and background colors."""


from dataclasses import dataclass
from typing import Optional

from .spaces import Color


@dataclass(frozen=True)
class ColorPair:
    """A color pair of foreground and background colors."""

    foreground: Optional[Color] = None
    background: Optional[Color] = None



"""Simple color conversions.

This module is a drop-in replacement for the
[standard `colorsys` module](https://docs.python.org/3/library/colorsys.html).

It provides extra functionality to the standard `colorsys` module, but also re-exports
its contents for convenience.

Examples
--------
>>> from ochre import colorsys

This module provides some conversions, among which is RGB from and to HCL:

>>> colorsys.rgb_to_hcl(0.2, 0.4, 0.4)  # doctest: +NUMBER
(3.4, 0.2, 0.4)
>>> colorsys.hcl_to_rgb(3.4, 0.2, 0.4)  # doctest: +NUMBER
(0.2, 0.4, 0.4)

For convenience, the module also re-exports the standard conversions from `colorsys`:

>>> colorsys.rgb_to_hsv(0.2, 0.4, 0.4)
(0.5, 0.5, 0.4)
>>> colorsys.hsv_to_rgb(0.5, 0.5, 0.4)
(0.2, 0.4, 0.4)
"""


from __future__ import annotations

import math
from colorsys import (
    hls_to_rgb,
    hsv_to_rgb,
    rgb_to_hls,
    rgb_to_hsv,
    rgb_to_yiq,
    yiq_to_rgb,
)
from typing import Text

from . import ansi256, web

__all__ = [
    # Implemented in this module
    "hcl_to_rgb",
    "rgb_to_hcl",
    # Re-exported from colorsys
    "hls_to_rgb",
    "hsv_to_rgb",
    "rgb_to_hls",
    "rgb_to_hsv",
    "rgb_to_yiq",
    "yiq_to_rgb",
]


def rgb_to_xyz(r: float, g: float, b: float) -> tuple[float, float, float]:
    """Convert the color from RGB coordinates to CIEXYZ coordinates."""
    if r > 0.04045:
        r = ((r + 0.055) / 1.055) ** 2.4
    else:
        r = r / 12.92
    if g > 0.04045:
        g = ((g + 0.055) / 1.055) ** 2.4
    else:
        g = g / 12.92
    if b > 0.04045:
        b = ((b + 0.055) / 1.055) ** 2.4
    else:
        b = b / 12.92

    x = 0.4124 * r + 0.3576 * g + 0.1805 * b
    y = 0.2126 * r + 0.7152 * g + 0.0722 * b
    z = 0.0193 * r + 0.1192 * g + 0.9505 * b
    return x, y, z


def xyz_to_rgb(x: float, y: float, z: float) -> tuple[float, float, float]:
    """Convert the color from CIEXYZ coordinates to RGB coordinates."""
    # We're using a higher precision matrix here see
    # <https://en.wikipedia.org/wiki/SRGB#sYCC_extended-gamut_transformation>
    r = 3.2406254 * x - 1.537208 * y - 0.4986286 * z
    g = -0.9689307 * x + 1.8757561 * y + 0.0415175 * z
    b = 0.0557101 * x - 0.2040211 * y + 1.0569959 * z

    if r > 0.0031308:
        r = 1.055 * (r ** (1 / 2.4)) - 0.055
    else:
        r = 12.92 * r
    if g > 0.0031308:
        g = 1.055 * (g ** (1 / 2.4)) - 0.055
    else:
        g = 12.92 * g
    if b > 0.0031308:
        b = 1.055 * (b ** (1 / 2.4)) - 0.055
    else:
        b = 12.92 * b

    return r, g, b


def luv_to_rgb(ell: float, u: float, v: float) -> tuple[float, float, float]:
    """Convert the color from CIELUV coordinates to RGB coordinates."""
    return xyz_to_rgb(*luv_to_xyz(ell, u, v))


def rgb_to_luv(r: float, g: float, b: float) -> tuple[float, float, float]:
    """Convert the color from RGB coordinates to CIELUV coordinates."""
    return xyz_to_luv(*rgb_to_xyz(r, g, b))


def hcl_to_rgb(h: float, c: float, ell: float) -> tuple[float, float, float]:
    """Convert the color from HCL coordinates to RGB coordinates."""
    return luv_to_rgb(*hcl_to_luv(h, c, ell))


def rgb_to_hcl(r: float, g: float, b: float) -> tuple[float, float, float]:
    """Convert the color from RGB coordinates to HCL coordinates."""
    return luv_to_hcl(*rgb_to_luv(r, g, b))


def xyz_to_luv(x: float, y: float, z: float) -> tuple[float, float, float]:
    """Convert the color from CIEXYZ coordinates to CIELUV coordinates."""
    u, v = _xyz_to_uv(x, y, z)

    if y > EPSILON:
        ell = 116 * y ** (1 / 3) - 16
    else:
        ell = KAPPA * y
    u, v = 13 * ell * (u - REF_UV_D65_2[0]), 13 * ell * (v - REF_UV_D65_2[1])

    return ell / 100, u / 100, v / 100


def luv_to_xyz(ell: float, u: float, v: float) -> tuple[float, float, float]:
    """Convert the color from CIELUV coordinates to CIEXYZ coordinates."""
    if ell == 0:
        return 0, 0, 0

    ell, u, v = 100 * ell, 100 * u, 100 * v
    u, v = _luv_to_uv(ell, u, v)

    four_v = 4 * v
    if ell > 8:
        y = ((ell + 16) / 116) ** 3
    else:
        y = ell / KAPPA
    x, z = 9 * y * u / four_v, y * (12 - 3 * u - 20 * v) / four_v

    return x, y, z


def luv_to_hcl(ell: float, u: float, v: float) -> tuple[float, float, float]:
    """Convert the color from CIELUV coordinates to HCL coordinates."""
    h = math.atan2(v, u)
    c = math.hypot(v, u)
    h = h + math.tau if h < 0 else h
    return h, c, ell


def hcl_to_luv(h: float, c: float, ell: float) -> tuple[float, float, float]:
    """Convert the color from HCL coordinates to CIELUV coordinates."""
    u = c * math.cos(h)
    v = c * math.sin(h)
    return ell, u, v


def rgb_to_hex(r: float, g: float, b: float) -> int:
    """Convert the color from RGB coordinates to hexadecimal."""
    return int(r * 255) << 16 | int(g * 255) << 8 | int(b * 255)


def hex_to_rgb(hc: int | Text) -> tuple[float, float, float]:
    """Convert the color from hexadecimal to RGB coordinates."""
    hc = hex_to_hex(hc)
    r = hc >> 16
    g = (hc >> 8) & 0xFF
    b = hc & 0xFF
    return r / 255, g / 255, b / 255


def web_color_to_hex(name: Text) -> int:
    """Convert the color from web color name to hexadecimal."""
    return hex_to_hex(web.colors[name])


def web_color_to_rgb(name: Text) -> tuple[float, float, float]:
    """Convert the color from web color name to RGB coordinates."""
    return hex_to_rgb(web_color_to_hex(name))


def ansi256_to_hex(c: int) -> int:
    """Convert the color from ANSI 256 color code to hexadecimal."""
    return hex_to_hex(ansi256.colors[c])


def ansi256_to_rgb(c: int) -> tuple[float, float, float]:
    """Convert the color from ANSI 256 color code to RGB coordinates."""
    return hex_to_rgb(ansi256_to_hex(c))


def hex_to_hex(hc: int | Text) -> int:
    """Ensure that the hexadecimal code is an integer."""
    if isinstance(hc, Text):
        return int(hc.lstrip("#"), 16)
    return hc


def _xyz_to_uv(x: float, y: float, z: float) -> tuple[float, float]:
    """Convert the color from CIEXYZ coordinates to uv chromaticity coordinates."""
    if x == y == 0:
        return 0, 0
    d = x + 15 * y + 3 * z
    return 4 * x / d, 9 * y / d


def _luv_to_uv(ell: float, u: float, v: float) -> tuple[float, float]:
    """Convert the color from CIELUV coordinates to uv chromaticity coordinates."""
    d = 13 * ell
    return u / d + REF_UV_D65_2[0], v / d + REF_UV_D65_2[1]


EPSILON = (6 / 29) ** 3
KAPPA = (29 / 3) ** 3


REF_XYZ_D65_2 = 0.95047, 1.00000, 1.08883
REF_UV_D65_2 = _xyz_to_uv(*REF_XYZ_D65_2)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ochre']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'ochre',
    'version': '0.4.0',
    'description': 'A tiny Python package for working with colors in a pragmatic way',
    'long_description': '[![PyPI](https://img.shields.io/pypi/v/ochre)](https://pypi.org/project/ochre/)\n[![Python package](https://github.com/getcuia/ochre/actions/workflows/python-package.yml/badge.svg)](https://github.com/getcuia/ochre/actions/workflows/python-package.yml)\n[![PyPI - License](https://img.shields.io/pypi/l/ochre)](https://github.com/getcuia/ochre/blob/main/LICENSE)\n\n# [ochre](https://github.com/getcuia/ochre#readme) 🏜️\n\n<div align="center">\n    <img class="hero" src="https://github.com/getcuia/ochre/raw/main/banner.jpg" alt="ochre" width="33%" />\n</div>\n\n> A down-to-earth approach to colors\n\nochre is a tiny Python package for working with colors in a pragmatic way. The\nfocus is on simplicity and ease of use, but also on human perception.\n\n## Features\n\n-   🎨 Focus on [RGB](https://en.wikipedia.org/wiki/RGB_color_model) and\n    [HCL](https://en.wikipedia.org/wiki/HCL_color_space) color spaces\n-   🖥️ [Web color names](https://en.wikipedia.org/wiki/Web_colors#Extended_colors)\n-   ♻️ Color conversions that easily integrate with the\n    [standard `colorsys` module](https://docs.python.org/3/library/colorsys.html)\n-   🗑️ Zero dependencies\n-   🐍 Python 3.8+\n\n## Installation\n\n```console\n$ pip install ochre\n```\n\n## Usage\n\n```python\nIn [1]: from ochre import Hex\n\nIn [2]: color = Hex("#CC7722")\n\nIn [3]: color.web_color\nOut[3]: WebColor(\'peru\')\n\nIn [4]: color = color.hcl\n\nIn [5]: color.hue\nOut[5]: 0.6373041934059377\n\nIn [6]: import math\n\nIn [7]: color.hue += math.radians(30)\n\nIn [8]: color.hue\nOut[8]: 1.1609029690042365\n\nIn [9]: color.web_color\nOut[9]: WebColor(\'goldenrod\')\n```\n\n## Credits\n\n[Photo](https://github.com/getcuia/ochre/raw/main/banner.jpg) by\n[Nicola Carter](https://unsplash.com/@ncarterwilts?utm_source=unsplash&utm_medium=referral&utm_content=creditCopyText)\non\n[Unsplash](https://unsplash.com/?utm_source=unsplash&utm_medium=referral&utm_content=creditCopyText).\n',
    'author': 'Felipe S. S. Schneider',
    'author_email': 'schneider.felipe@posgrad.ufsc.br',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/getcuia/ochre',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
